catch<-function(tag, expr) {
    "__CATCH_TAG__" <- as.character(tag)
    "__CATCH_THROWER__"<- make.thrower(return(get("__CATCH_VALUE__")))
    "__CATCH_VALUE__" <- NULL
    expr
}

make.thrower <- function(expr) function() expr
throw<-function(tag, expr, no.tag = tag.not.found) {
    value <- expr # forces evaluation of expr
    tag <- as.character(tag)
    env <- get.target.frame(tag, "__CATCH_TAG__")
    assign("__CATCH_VALUE__", value, env = env)
    fun <- get("__CATCH_THROWER__", env = env)
    fun()
}

##****parent.env with arg > 1 doesn't seem to work as advertized??
get.target.frame <- function(tag, name) {
    n <- sys.nframe()
    if (n > 1)
        for (i in (n-1):1) {
            env <- sys.frame(i)
            if (exists(name, env = env) &&
                get(name, env = env) == tag)
                return(env)
        }
    stop(paste("no catch for tag \"", tag, "\"", sep = ""))
} 
dynamic.exists <- function(name) {
    n <- sys.nframe()
    if (n > 1)
        for (i in (n-1):1)
            if (exists(name, env = sys.frame(i)))
                return(TRUE)
    FALSE;
}
print.exception <- function(e, ...)
    cat("<", class(e)[1], ": ", as.character(e), ">\n", sep="")
as.character.exception <- function(e, ...) {
    warning(paste("no as.character method for exception of class", class(e)))
    as.character(unclass(e))
}
simple.exception <- function(string) {
    class <- c("simple.exception", "exception")
    structure(list(string=as.character(string)), class=class)
}

as.character.simple.exception <- function(e, ...)
    e$string
default.exception.handler <- function(e)
    UseMethod("default.exception.handler", e)

default.exception.handler.exception <- function(e)
    stop(paste("unhandled exception (", class(e)[1],
               "): ", as.character(e), sep = ""))
raise.exception <- function(e) {
    reset.error.options() # turn off to avoid recursion
    if (! inherits(e, "exception"))
        e <- simple.exception(as.character(e))
    if (exists.try.catch())
        throw("__TRY_CATCH__", list(value = e, throw = TRUE))
    else
        default.exception.handler(e)
}

exists.try.catch <- function() dynamic.exists("__TRY_CATCH__")
try.catch <- function(expr, ..., finally = NULL) {
    op <- set.error.options()
    on.exit({restore.error.options(op); finally })
    result <- local({
        "__TRY_CATCH__" <- NULL
        catch("__TRY_CATCH__", list(value = expr, throw = FALSE))
    })
    if (result$throw) {
        set.error.options() # turn back on because turned off in raise
        .Call("ResetErrorHandling") # is this the right place for this?
        handlers <- list(...)
        names <- names(handlers)
        e <- result$value
        for (i in seq(along = names))
            if (inherits(e, names[i]))
                return(handlers[[i]](e))
        on.exit()                  # to get
        restore.error.options(op)  # things reasonable
        finally                    # for stop calls going to default handler
        raise.exception(e)
    }
    else result$value
}
error.converter <- function() {
    raise.exception(simple.exception(geterrmessage()))
}
set.error.options <- function() {
    op <- options()
    options(show.error.messages = FALSE)
    options(error = quote(error.converter()))
    if (is.logical(op$show.error.messages)) show <- op$show.error.messages
    else show <- TRUE
    if (is.logical(op$error)) err <- NULL
    else err <- op$error
    list(show.error.messages = show, error = err)
    
}
reset.error.options <- function() {
    options(show.error.messages = TRUE)
    options(error = NULL)
}
restore.error.options <- function(op) {
    options(op)
}
.First.lib <- function(lib, pkg)
    library.dynam(pkg, pkg, lib)
